"""Unit tests for schedule.py"""
import datetime
import functools
import mock
import unittest
import time
# Silence "missing docstring", "method could be a function",
# "class already defined", and "too many public methods" messages:
# pylint: disable-msg=R0201,C0111,E0102,R0904,R0901

import schedule
from schedule import every


class Time(object):
    pass


def to_time(t):
    dt = datetime.datetime.fromtimestamp(t)
    return dt


def make_mock_job(name=None):
    job = mock.Mock()
    job.__name__ = name or 'job'
    return job


class mock_datetime(object):
    """
    Monkey-patch datetime for predictable results
    """

    def __init__(self, year, month, day, hour, minute):
        self.year = year
        self.month = month
        self.day = day
        self.hour = hour
        self.minute = minute

    def __enter__(self):
        def mock_time():
            return time.mktime(datetime.datetime(
                self.year, self.month, self.day, self.hour, self.minute)
                .timetuple())

        self.original_time = time.time
        time.time = mock_time

    def __exit__(self, *args, **kwargs):
        time.time = self.original_time


class SchedulerTests(unittest.TestCase):

    def setUp(self):
        schedule.clear()

    def test_time_units(self):
        assert every().seconds.unit == 'seconds'
        assert every().minutes.unit == 'minutes'
        assert every().hours.unit == 'hours'
        assert every().days.unit == 'days'
        assert every().weeks.unit == 'weeks'

    def test_singular_time_units_match_plural_units(self):
        assert every().second.unit == every().seconds.unit
        assert every().minute.unit == every().minutes.unit
        assert every().hour.unit == every().hours.unit
        assert every().day.unit == every().days.unit
        assert every().week.unit == every().weeks.unit

    def test_at_time(self):
        mock_job = make_mock_job()
        assert to_time(every().day.at('10:30').do(
            mock_job).next_run).hour == 10
        assert to_time(every().day.at('10:30').do(
            mock_job).next_run).minute == 30

    def test_at_time_hour(self):
        with mock_datetime(2010, 1, 6, 12, 20):
            mock_job = make_mock_job()
            assert to_time(every().hour.at(':30').do(
                mock_job).next_run).hour == 12
            assert to_time(every().hour.at(':30').do(
                mock_job).next_run).minute == 30
            assert to_time(every().hour.at(':10').do(
                mock_job).next_run).hour == 13
            assert to_time(every().hour.at(':10').do(
                mock_job).next_run).minute == 10
            assert to_time(every().hour.at(':00').do(
                mock_job).next_run).hour == 13
            assert to_time(every().hour.at(':00').do(
                mock_job).next_run).minute == 0

    def test_next_run_time(self):
        with mock_datetime(2010, 1, 6, 12, 15):
            mock_job = make_mock_job()
            assert schedule.next_run() is None
            assert to_time(every().minute.do(mock_job).next_run).minute == 16
            assert to_time(every(5).minutes.do(mock_job).next_run).minute == 20
            assert to_time(every().hour.do(mock_job).next_run).hour == 13
            assert to_time(every().day.do(mock_job).next_run).day == 7
            assert to_time(every().day.at('09:00').do(
                mock_job).next_run).day == 7
            assert to_time(every().day.at('12:30').do(
                mock_job).next_run).day == 6
            assert to_time(every().week.do(mock_job).next_run).day == 13
            assert to_time(every().monday.do(mock_job).next_run).day == 11
            assert to_time(every().tuesday.do(mock_job).next_run).day == 12
            assert to_time(every().wednesday.do(mock_job).next_run).day == 13
            assert to_time(every().thursday.do(mock_job).next_run).day == 7
            assert to_time(every().friday.do(mock_job).next_run).day == 8
            assert to_time(every().saturday.do(mock_job).next_run).day == 9
            assert to_time(every().sunday.do(mock_job).next_run).day == 10

    def test_run_all(self):
        mock_job = make_mock_job()
        every().minute.do(mock_job)
        every().hour.do(mock_job)
        every().day.at('11:00').do(mock_job)
        schedule.run_all()
        assert mock_job.call_count == 3

    def test_job_func_args_are_passed_on(self):
        mock_job = make_mock_job()
        every().second.do(mock_job, 1, 2, 'three', foo=23, bar={})
        schedule.run_all()
        mock_job.assert_called_once_with(1, 2, 'three', foo=23, bar={})

    # def test_to_string(self):
    #    def job_fun():
    #        pass
    #    s = str(every().minute.do(job_fun, 'foo', bar=23))
    #    assert 'job_fun' in s
    #    assert 'foo' in s
    #    assert 'bar=23' in s

    # def test_to_string_lambda_job_func(self):
    #    assert len(str(every().minute.do(lambda: 1))) > 1
    #    assert len(str(every().day.at('10:30').do(lambda: 1))) > 1

    # def test_to_string_functools_partial_job_func(self):
    #    def job_fun(arg):
    #        pass
    #    job_fun = functools.partial(job_fun, 'foo')
    #    job_repr = repr(every().minute.do(job_fun, bar=True, somekey=23))
    #    assert 'functools.partial' in job_repr
    #    assert 'bar=True' in job_repr
    #    assert 'somekey=23' in job_repr

    def test_run_pending(self):
        """Check that run_pending() runs pending jobs.
        We do this by overriding datetime.datetime with mock objects
        that represent increasing system times.

        Please note that it is *intended behavior that run_pending() does not
        run missed jobs*. For example, if you've registered a job that
        should run every minute and you only call run_pending() in one hour
        increments then your job won't be run 60 times in between but
        only once.
        """
        mock_job = make_mock_job()

        with mock_datetime(2010, 1, 6, 12, 15):
            every().minute.do(mock_job)
            every().hour.do(mock_job)
            every().day.do(mock_job)
            every().sunday.do(mock_job)
            schedule.run_pending()
            assert mock_job.call_count == 0

        with mock_datetime(2010, 1, 6, 12, 16):
            schedule.run_pending()
            assert mock_job.call_count == 1

        with mock_datetime(2010, 1, 6, 13, 16):
            mock_job.reset_mock()
            schedule.run_pending()
            assert mock_job.call_count == 2

        with mock_datetime(2010, 1, 7, 13, 16):
            mock_job.reset_mock()
            schedule.run_pending()
            assert mock_job.call_count == 3

        with mock_datetime(2010, 1, 10, 13, 16):
            mock_job.reset_mock()
            schedule.run_pending()
            assert mock_job.call_count == 4

    def test_run_every_weekday_at_specific_time_today(self):
        mock_job = make_mock_job()
        with mock_datetime(2010, 1, 6, 13, 16):
            every().wednesday.at('14:12').do(mock_job)
            schedule.run_pending()
            assert mock_job.call_count == 0

        with mock_datetime(2010, 1, 6, 14, 16):
            schedule.run_pending()
            assert mock_job.call_count == 1

    def test_run_every_weekday_at_specific_time_past_today(self):
        mock_job = make_mock_job()
        with mock_datetime(2010, 1, 6, 13, 16):
            every().wednesday.at('13:15').do(mock_job)
            schedule.run_pending()
            assert mock_job.call_count == 0

        with mock_datetime(2010, 1, 13, 13, 14):
            schedule.run_pending()
            assert mock_job.call_count == 0

        with mock_datetime(2010, 1, 13, 13, 16):
            schedule.run_pending()
            assert mock_job.call_count == 1

    def test_run_every_n_days_at_specific_time(self):
        mock_job = make_mock_job()
        with mock_datetime(2010, 1, 6, 11, 29):
            every(2).days.at('11:30').do(mock_job)
            schedule.run_pending()
            assert mock_job.call_count == 0

        with mock_datetime(2010, 1, 6, 11, 31):
            schedule.run_pending()
            assert mock_job.call_count == 0

        with mock_datetime(2010, 1, 7, 11, 31):
            schedule.run_pending()
            assert mock_job.call_count == 0

        with mock_datetime(2010, 1, 8, 11, 29):
            schedule.run_pending()
            assert mock_job.call_count == 0

        with mock_datetime(2010, 1, 8, 11, 31):
            schedule.run_pending()
            assert mock_job.call_count == 1

        with mock_datetime(2010, 1, 10, 11, 31):
            schedule.run_pending()
            assert mock_job.call_count == 2

    def test_next_run_property(self):
        original_datetime = datetime.datetime
        with mock_datetime(2010, 1, 6, 13, 16):
            hourly_job = make_mock_job('hourly')
            daily_job = make_mock_job('daily')
            every().day.do(daily_job)
            every().hour.do(hourly_job)
            assert len(schedule.jobs) == 2
            # Make sure the hourly job is first
            assert to_time(schedule.next_run()) == original_datetime(
                2010, 1, 6, 14, 16)
            assert schedule.idle_seconds() == 60 * 60

    def test_cancel_job(self):
        def stop_job():
            return schedule.CancelJob
        mock_job = make_mock_job()

        every().second.do(stop_job)
        mj = every().second.do(mock_job)
        assert len(schedule.jobs) == 2

        schedule.run_all()
        assert len(schedule.jobs) == 1
        assert schedule.jobs[0] == mj

        schedule.cancel_job('Not a job')
        assert len(schedule.jobs) == 1
        schedule.default_scheduler.cancel_job('Not a job')
        assert len(schedule.jobs) == 1

        schedule.cancel_job(mj)
        assert len(schedule.jobs) == 0

    def test_functools_partial_job_func(self):
        def job_fun(arg):
            pass
        job_fun = functools.partial(job_fun, 'foo')
        every().minute.do(job_fun, bar=True, somekey=23)
